""" Library for download and parse financial data from various sources.
"""

__author__ = "Stanley Sane"
__copyright__ = "Copyright (c) 2018-2021, Sane Development"
__license__ = "The BSD 3-Clause License"

__version__ = '1.0'
